# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Tools

from importlib.resources import files


def get_verilog():
    """
    Get the path to the Verilog source files for the APB agent.

    :return: Path to the Verilog source files
    :rtype: pathlib.Path
    """
    path = files("avl_apb.rtl").joinpath("avl_apb.sv")
    print(path)



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Base Driver


import asyncio

import avl
import cocotb
from cocotb.triggers import FallingEdge, First

from ._item import SequenceItem


class Driver(avl.Driver):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the Driver for the APB agent.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        self.i_f = avl.Factory.get_variable(f"{self.get_full_name()}.i_f", None)
        self.rate_limit = avl.Factory.get_variable(f"{self.get_full_name()}.rate_limit", lambda : 1.0)
        """Rate limit for driving signals. lambda function (0.0 - 1.0)"""

        if not callable(self.rate_limit):
            raise TypeError("rate_limit must be a callable (lambda function) that returns a float between 0.0 and 1.0")

    async def reset(self) -> None:
        """
        Reset the driver by setting all signals to their default values.
        This method is called when the driver is reset.

        By default 0's all signals - can be overridden in subclasses to add randomization or other behavior.
        """

        raise NotImplementedError("Reset method must be implemented in subclasses")

    async def wait_on_reset(self) -> None:
        """
        Wait for the reset signal to go low and then call the reset method.
        This method is called to ensure that the driver is reset before driving any signals.
        It waits for the presetn signal to go low, indicating that the reset is active,
        and then calls the reset method to set all signals to their default values.
        """

        try:
            await FallingEdge(self.i_f.presetn)
            await self.reset()
        except asyncio.CancelledError:
            raise
        except Exception:
            pass

    async def quiesce(self) -> None:
        """
        Quiesce the driver by setting the psel signal to 0.
        This method is called when the driver is quiesced.

        By default calls reset() to set all signals to their default values.
        Can be overridden in subclasses to add randomization or other behavior.
        """

        await self.reset()

    async def drive(self, item : SequenceItem) -> None:
        """
        Drive the signals based on the provided sequence item.
        This method is called to drive the signals of the APB interface.

        :param item: The sequence item containing the values to drive
        :type item: SequenceItem
        """
        raise NotImplementedError("Drive method must be implemented in subclasses")

    async def get_next_item(self, item : SequenceItem = None) -> SequenceItem:
        """
        Get the next sequence item.

        For the Request driver this method retrieves the next sequence item from the sequencer or
        the previously reset interrupted item.

        The implementation ensures items are driven on the rising edge of pclk, when not in reset,
        while allowing for back-to-back requests if the sequencer provides them.

        For the completion driver this method adjusts the completion side of the observed request.

        :param item: The sequence item to retrieve, defaults to None
        :type item: SequenceItem, optional
        :return: The next sequence item
        :rtype: SequenceItem
        :raises NotImplementedError: If the method is not implemented in subclasses
        """

        raise NotImplementedError("get_next_item method must be implemented in subclasses")

    async def run_phase(self):
        """
        Run phase for the Requester Driver.
        This method is called during the run phase of the simulation.
        It is responsible for driving the request signals based on the sequencer's items.

        :raises NotImplementedError: If the run phase is not implemented.
        """
        item = None

        # Start from reset state
        await self.reset()

        while True:
            item = await self.get_next_item(item)

            drive_task = cocotb.start_soon(self.drive(item))
            reset_task = cocotb.start_soon(self.wait_on_reset())

            await First(drive_task, reset_task)

            if drive_task.done():
                item = None

            for t in [drive_task, reset_task]:
                if not t.done():
                    t.cancel()

__all__ = ["Driver"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Sequence Item

from typing import Any

import avl
from z3 import BoolRef, Implies


class SequenceItem(avl.SequenceItem):
    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the sequence item

        :param name: Name of the sequence item
        :param parent: Parent component of the sequence item
        """
        super().__init__(name, parent)

        # Handle to interface - defines capabilities and parameters
        i_f = avl.Factory.get_variable(f"{self.get_full_name()}.i_f", None)

        self.paddr = avl.Logic(0, width=len(i_f.paddr), fmt=hex)
        """Address"""

        self.psel = avl.Logic(0, width=len(i_f.psel), fmt=hex)
        """Select (1-hot)"""

        self.pwrite = avl.Logic(0, width=len(i_f.pwrite), fmt=str)
        """Write enable"""

        self.pwdata = avl.Logic(0, width=len(i_f.pwdata), fmt=hex)
        """Write data"""

        self.prdata = avl.Logic(0, width=len(i_f.prdata), fmt=hex)
        """Read data"""

        if hasattr(i_f, "pslverr"):
            self.pslverr = avl.Logic(0, width=len(i_f.pslverr), fmt=str)
            """Slave error (>= version 3)"""

        if hasattr(i_f, "pstrb"):
            self.pstrb = avl.Logic(0, width=len(i_f.pstrb), fmt=hex)
            """Write strobe (byte enable) (>= version 4)"""

        if hasattr(i_f, "pprot"):
            self.pprot = avl.Logic(0, width=len(i_f.pprot), fmt=hex)
            """Protection bits (optional >= version 4)"""

        if hasattr(i_f, "pnse"):
            self.pnse = avl.Logic(0, width=len(i_f.pnse), fmt=str)
            """Non-secure enable (optional >= version 5)"""

        if hasattr(i_f, "pwakeup"):
            self.goto_sleep = avl.Logic(0, width=len(i_f.pwakeup), fmt=str)
            """Wakeup indication (optional >= version 5)"""

        if hasattr(i_f, "pauser"):
            self.pauser = avl.Logic(0, width=len(i_f.pauser), fmt=hex)
            """User Request Sideband (optional >= version 5)"""

        if hasattr(i_f, "pwuser"):
            self.pwuser = avl.Logic(0, width=len(i_f.pwuser), fmt=hex)
            """User Write Sideband (optional >= version 5)"""

        if hasattr(i_f, "pruser"):
            self.pruser = avl.Logic(0, width=len(i_f.pruser), fmt=hex)
            """User Read Sideband (optional >= version 5)"""

        if hasattr(i_f, "pbuser"):
            self.pbuser = avl.Logic(0, width=len(i_f.pbuser), fmt=hex)
            """User Response Sideband (optional >= version 5)"""

        # Constraints
        self.add_constraint("c_psel_valid", lambda x : x != 0, self.psel)
        self.add_constraint("c_psel_1hot", lambda x : (x & (x-1) == 0), self.psel)

        if hasattr(self, "pstrb"):
            self.add_constraint("c_pstrb", lambda x, y : Implies(x == 0, y == 0), self.pwrite, self.pstrb)

        # Monitor only attributes used for debug and coverage
        if hasattr(i_f, "pready"):
            self.wait_cycles = 0
            """Wait cycles - cycles from enable to ready (monitor only)"""
            self.set_field_attributes("wait_cycles", compare=False)


        if hasattr(i_f, "pwakeup"):
            self.time_since_wakeup = 0
            """Time since last wakeup - used for debug and coverage (monitor only)"""
            self.set_field_attributes("time_since_wakeup", compare=False)

        # By default transpose to make more readable
        self.set_table_fmt(transpose=True)

    def set(self, name : str, value : int) -> None:
        """
        Set the value of a field in the sequence item - if it exists.

        :param name: Name of the field to set
        :param value: Value to set for the field
        """
        signal = getattr(self, name, None)
        if signal is not None:
            signal.value = value

    def get(self, name : str, default : Any = None) -> int:
        """
        Get the value of a field in the sequence item - if it exists.

        :param name: Name of the field to get
        :param default: Default value to return if the field does not exist
        :return: Value of the field or default value
        """
        signal = getattr(self, name, None)
        if signal is not None:
            return signal.value
        return default

    def randomize_request(self, hard: list[BoolRef] = None, soft: list[BoolRef] = None) -> None:
        """
        Randomize the request fields of the sequence item.
        """
        for f in ["prdata", "pslverr", "pruser", "pbuser"]:
            if hasattr(self, f):
                getattr(self, f)._auto_random_ = False

        self.randomize(hard=hard, soft=soft)

        for f in ["prdata", "pslverr", "pruser", "pbuser"]:
            if hasattr(self, f):
                getattr(self, f)._auto_random_ = True

    def randomize_completion(self, hard: list[BoolRef] = None, soft: list[BoolRef] = None) -> None:
        """
        Randomize the completion fields of the sequence item.
        """

        for f in ["paddr", "psel", "pwrite", "pwdata", "pstrb", "pprot", "pnse", "pauser", "pwuser", "goto_sleep"]:
            if hasattr(self, f):
                getattr(self, f)._auto_random_ = False

        self.randomize(hard=hard, soft=soft)

        for f in ["paddr", "psel", "pwrite", "pwdata", "pstrb", "pprot", "pnse", "pauser", "pwuser", "goto_sleep"]:
            if hasattr(self, f):
                getattr(self, f)._auto_random_ = True

__all__ = ["SequenceItem"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Driver

import asyncio
import random

import avl
import cocotb
from cocotb.triggers import First, NextTimeStep, RisingEdge

from ._driver import Driver
from ._item import SequenceItem


class ReqDriver(Driver):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the Requester Driver for the AMBA agent.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        # Pre and Post Wakeup
        if hasattr(self.i_f, "pwakeup"):
            self.pre_wakeup =  avl.Factory.get_variable(f"{self.get_full_name()}.pre_wakeup", lambda : 0.1)
            """Pre-wakeup delay - time to wait before driving the wakeup signal (0.0 - 1.0) (>= version 5)"""
            self.post_wakeup = avl.Factory.get_variable(f"{self.get_full_name()}.post_wakeup", lambda : 0.1)
            """Post-wakeup delay - time to wait after driving the wakeup signal (0.0 - 1.0) (>= version 5)"""

            if not callable(self.pre_wakeup) or not callable(self.post_wakeup):
                raise TypeError("pre_wakeup and post_wakeup must be callable (lambda functions) that return a float between 0.0 and 1.0")

    async def reset(self) -> None:
        """
        Reset the driver by setting all signals to their default values.
        This method is called when the driver is reset.

        By default 0's all signals - can be overridden in subclasses to add randomization or other behavior.
        """

        self.i_f.set("pwakeup", 0)
        self.i_f.set("paddr", 0)
        self.i_f.set("psel", 0)
        self.i_f.set("penable", 0)
        self.i_f.set("pwrite", 0)
        self.i_f.set("pwdata", 0)
        self.i_f.set("pstrb", 0)
        self.i_f.set("pprot", 0)
        self.i_f.set("pnse", 0)
        self.i_f.set("pwakeup", 0)
        self.i_f.set("pauser", 0)
        self.i_f.set("pwuser", 0)

    async def quiesce(self) -> None:
        """
        Quiesce the driver by setting the psel signal to 0.
        This method is called when the driver is quiesced.

        By default calls reset() to set all signals to their default values.
        Can be overridden in subclasses to add randomization or other behavior.
        """

        self.i_f.set("paddr", 0)
        self.i_f.set("psel", 0)
        self.i_f.set("penable", 0)
        self.i_f.set("pwrite", 0)
        self.i_f.set("pwdata", 0)
        self.i_f.set("pstrb", 0)
        self.i_f.set("pprot", 0)
        self.i_f.set("pnse", 0)
        self.i_f.set("pauser", 0)
        self.i_f.set("pwuser", 0)

    async def drive(self, item : SequenceItem) -> None:
        """
        Drive the signals based on the provided sequence item.
        This method is called to drive the signals of the AMBA interface.

        :param item: The sequence item containing the values to drive
        :type item: SequenceItem
        """
        awake = False
        try:
            self.i_f.set("penable", 0)

            # Rate Limiter
            rate = self.rate_limit()
            while random.random() > rate:
                await RisingEdge(self.i_f.pclk)

            if hasattr(self.i_f, "pwakeup") and not awake:
                self.i_f.set("pwakeup", 1)
                delay = self.pre_wakeup()
                while random.random() > delay:
                    await RisingEdge(self.i_f.pclk)


            self.i_f.set("psel",   item.get("psel"))
            self.i_f.set("paddr",  item.get("paddr"))
            self.i_f.set("pwrite", item.get("pwrite"))
            self.i_f.set("pwdata", item.get("pwdata"))
            self.i_f.set("pstrb",  item.get("pstrb"))
            self.i_f.set("pprot",  item.get("pprot"))
            self.i_f.set("pnse",   item.get("pnse"))
            self.i_f.set("pauser", item.get("pauser"))
            self.i_f.set("pwuser", item.get("pwuser"))

            await RisingEdge(self.i_f.pclk)
            self.i_f.set("penable", 1)

            while True:
                await RisingEdge(self.i_f.pclk)
                if self.i_f.get("pready", 1) and self.i_f.get("pwakeup", 1):
                    break

            item.set("prdata", self.i_f.get("prdata"))
            item.set("pruser", self.i_f.get("pruser"))
            item.set("pslverr", self.i_f.get("pslverr"))
            item.set("pbuser", self.i_f.get("pbuser"))

            # Clear the bus
            await self.quiesce()

            # Post wakeup
            if item.get("goto_sleep", False):
                delay = self.post_wakeup()
                while random.random() > delay:
                    await RisingEdge(self.i_f.pclk)
                self.i_f.set("pwakeup", 0)
                awake = False
            else:
                awake = True

            item.set_event("done")

        except asyncio.CancelledError:
            raise
        except Exception:
            self.warning(f"Requester drive task for item was cancelled by reset:\n{item}")

    async def get_next_item(self, item : SequenceItem = None) -> SequenceItem:
        """
        Get the next sequence item.

        This method retrieves the next sequence item from the sequencer or
        the previously reset interrupted item.

        The implementation ensures items are driven on the rising edge of pclk, when not in reset,
        while allowing for back-to-back requests if the sequencer provides them.

        :param item: The sequence item to retrieve, defaults to None
        :type item: SequenceItem, optional
        :return: The next sequence item
        :rtype: SequenceItem
        :raises NotImplementedError: If the method is not implemented in subclasses
        """

        async def next_time_step() -> None:
            await NextTimeStep()

        if item is not None:
            next_item = item
        else:
            a = cocotb.start_soon(self.seq_item_port.blocking_get())
            b = cocotb.start_soon(next_time_step())

            await First(a, b)
            if not a.done():
                await a
                await RisingEdge(self.i_f.pclk)
            next_item = a.result()

        while self.i_f.get("presetn") == 0:
            await RisingEdge(self.i_f.pclk)

        return next_item

__all__ = ["ReqDriver"]



from ._agent import Agent
from ._agent_cfg import AgentCfg
from ._bandwidth import Bandwidth
from ._cdriver import CplDriver, CplMemoryDriver, CplRandomDriver
from ._coverage import Coverage
from ._item import SequenceItem
from ._monitor import Monitor
from ._rdriver import ReqDriver
from ._rsequence import ReqSequence

# Add version
__version__: str = "0.2.0"

__all__ = [
    "Agent",
    "AgentCfg",
    "Bandwidth",
    "CplDriver",
    "CplMemoryDriver",
    "CplRandomDriver",
    "Coverage",
    "SequenceItem",
    "Monitor",
    "ReqDriver",
    "ReqSequence",
]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Requester Sequence

import random

import avl
from z3 import And

from ._item import SequenceItem


class ReqSequence(avl.Sequence):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the sequence item

        :param name: Name of the sequence item
        :param parent: Parent component of the sequence item
        """
        super().__init__(name, parent)

        self.i_f = avl.Factory.get_variable(f"{self.get_full_name()}.i_f", None)
        """Handle to interface - defines capabilities and parameters"""

        self.n_items = avl.Factory.get_variable(f"{self.get_full_name()}.n_items", 1)
        """Number of items in the sequence (default 1)"""

        self.ranges = avl.Factory.get_variable(f"{self.get_full_name()}.ranges", None)
        """List of ranges for the address space (optional, default None)"""

    async def body(self) -> None:
        """
        Body of the sequence
        """

        self.info(f"Starting sequence {self.get_full_name()} with {self.n_items} items")
        for _ in range(self.n_items):
            item = SequenceItem(f"from_{self.name}", self)
            await self.start_item(item)
            if self.ranges is not None:
                (psel, lo, hi) = random.choices(list(self.ranges.keys()), weights=list(self.ranges.values()), k=1)[0]

                item.add_constraint("_c_psel_", lambda x,y=psel: x == int(1 << y), item.psel)
                item.add_constraint("_c_paddr_", lambda x,y=lo,z=(hi-self.i_f.PSTRB_WIDTH): And(x >= y, x <= z), item.paddr)

            item.randomize_request()
            await self.finish_item(item)

__all__ = ["ReqSequence"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Bandwidth Monitor

import avl
import cocotb
import matplotlib.pyplot as plt
import numpy as np
from cocotb.triggers import Timer
from cocotb.utils import get_sim_time


class Bandwidth(avl.Component):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize Bandwidth Monitor

        Class contains variable window_ns which defines the time window in nanoseconds. This
        is configurable via the factory.

        :param name: Name of the bandwidth monitor.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)
        self.i_f = avl.Factory.get_variable(f"{self.get_full_name()}.i_f", None)

        self.window_ns = avl.Factory.get_variable(f"{self.get_full_name()}.window_ns", 1000)
        """ Rolling ime window in nanoseconds for bandwidth calculation. Default is 1000 ns. """

        self.item_port = avl.List()
        self.bandwidth = {}

    async def run_phase(self) -> None:
        """
        Run phase for the coverage component.

        """
        async def counter():
            while True:
                self.item = await self.item_port.blocking_get()

                incr = self.i_f.PSTRB_WIDTH
                if self.item.get("pwrite"):
                    pstrb = self.item.get("pstrb", None)
                    if pstrb is not None:
                        incr = bin(pstrb).count("1")
                self.bandwidth[t] += incr

        cocotb.start_soon(counter())

        while True:
            t = get_sim_time(unit="ns")
            self.bandwidth[t] = 0

            self.raise_objection()
            await Timer(self.window_ns, unit="ns")
            self.drop_objection()

    async def report_phase(self) -> None:
        """
        Report phase for the bandwidth component.

        Generate plot of bytes on bus over time windows

        """
        times = list(self.bandwidth.keys())
        counts = list(self.bandwidth.values())

        if len(times) > 1:
            # For unevenly spaced data, calculate individual widths
            # This uses the minimum spacing to avoid overlapping bars
            spacings = np.diff(times)
            bar_width = np.min(spacings)
        else:
            # For single bar, use a default width
            bar_width = 1

        # Create plot
        plt.figure(figsize=(8, 4))
        plt.bar(times, counts, width=bar_width, align='center')
        plt.title("Bus Bandwidth (Bytes over Time)")
        plt.xlabel("Time (ns)")
        plt.ylabel("Byte Count")
        plt.grid(True)

        # Export to PNG
        plt.savefig(f"{self.get_full_name()}.png")
        plt.close()

__all__ = ["Bandwidth"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Monitor

import asyncio

import avl
import cocotb
from cocotb.triggers import FallingEdge, First, RisingEdge
from cocotb.utils import get_sim_time

from ._item import SequenceItem


class Monitor(avl.Monitor):
    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the AMBA Monitor for the APB agent.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        self.i_f = avl.Factory.get_variable(f"{self.get_full_name()}.i_f", None)

        self.wakeup = 0

    async def monitor(self) -> None:
        """
        Monitor the APB bus signals and create sequence items based on the activity.
        This method is called to monitor the bus signals and create sequence items
        when there is activity on the bus.
        """
        try:
            item = SequenceItem(f"from_{self.name}", self)
            item.wait_cycles = 0

            while bool(self.i_f.get("penable")):
                await RisingEdge(self.i_f.pclk)

            item.time_since_wakeup = get_sim_time("ns") - self.wakeup

            item.set("psel",   self.i_f.get("psel"))
            item.set("paddr",  self.i_f.get("paddr"))
            item.set("pwrite", self.i_f.get("pwrite"))
            item.set("pwdata", self.i_f.get("pwdata"))
            item.set("pstrb" , self.i_f.get("pstrb"))
            item.set("pprot",  self.i_f.get("pprot"))
            item.set("pnse",   self.i_f.get("pnse"))
            item.set("pauser", self.i_f.get("pauser"))
            item.set("pwuser", self.i_f.get("pwuser"))

            while True:
                await RisingEdge(self.i_f.pclk)
                if bool(self.i_f.get("pready", 1)):
                    break
                item.wait_cycles += 1

            item.set("prdata", self.i_f.get("prdata"))
            item.set("pslverr", self.i_f.get("pslverr"))
            item.set("prdata", self.i_f.get("prdata"))
            item.set("pbuser", self.i_f.get("pbuser"))

            # Send to export
            self.item_export.write(item)

        except asyncio.CancelledError:
            raise
        except Exception:
            self.debug(f"Drive task for item {item} was cancelled by reset")
            item.set_event("done")

    async def run_phase(self):
        """
        Run phase for the Requester Driver.
        This method is called during the run phase of the simulation.
        It is responsible for driving the request signals based on the sequencer's items.

        :raises NotImplementedError: If the run phase is not implemented.
        """

        async def wait_on_wakeup() -> None:
            while True:
                await RisingEdge(self.i_f.pwakeup)
                self.wakeup = get_sim_time("ns")

        async def wait_on_reset() -> None:
            try:
                await FallingEdge(self.i_f.presetn)
                await self.reset()
            except asyncio.CancelledError:
                raise
            except Exception:
                pass

        # Start Wakeup Monitor
        if hasattr(self.i_f, "pwakeup"):
            cocotb.start_soon(wait_on_wakeup())

        while True:
            if self.i_f.presetn == 0 or self.i_f.psel == 0 or self.i_f.get("pwakeup", 1) == 0:
                await RisingEdge(self.i_f.pclk)
                continue

            monitor_task = cocotb.start_soon(self.monitor())
            reset_task = cocotb.start_soon(wait_on_reset())

            await First(monitor_task, reset_task)
            for t in [monitor_task, reset_task]:
                if not t.done():
                    t.cancel()

__all__ = ["Monitor"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Driver

import asyncio
import random

import avl
from cocotb.triggers import RisingEdge

from ._driver import Driver
from ._item import SequenceItem


class CplDriver(Driver):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the Requester Driver for the AMBA agent.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        self.idx = avl.Factory.get_variable(f"{self.get_full_name()}.idx", 0)
        """Index of the driver in the AMBA interface (psel), used to select the correct signals."""

    async def reset(self) -> None:
        """
        Reset the driver by setting all signals to their default values.
        This method is called when the driver is reset.

        By default 0's all signals - can be overridden in subclasses to add randomization or other behavior.
        """

        self.i_f.set("prdata", 0)
        self.i_f.set("pready", 0)
        self.i_f.set("pslverr", 0)
        self.i_f.set("pbuser", 0)
        self.i_f.set("pruser", 0)

    async def drive(self, item : SequenceItem) -> None:
        """
        Drive the signals based on the provided sequence item.
        This method is called to drive the signals of the AMBA interface.

        :param item: The sequence item containing the values to drive
        :type item: SequenceItem
        """
        try:

            if self.i_f.get("pready") is not None:
                self.i_f.set("pready", 0)
                # Rate Limiter
                rate = self.rate_limit()
                while random.random() > rate:
                    await RisingEdge(self.i_f.pclk)
                self.i_f.set("pready", 1)

            self.i_f.set("prdata", item.get("prdata"))
            self.i_f.set("pslverr", item.get("pslverr"))
            self.i_f.set("pruser", item.get("pruser"))
            self.i_f.set("pbuser", item.get("pbuser"))

            # Wait for the next clock edge to drive the item
            await RisingEdge(self.i_f.pclk)

            # Clear the bus
            await self.quiesce()

        except asyncio.CancelledError:
            raise
        except Exception:
            self.debug(f"Completer drive task for item was cancelled by reset:\n{item}")

    async def get_next_item(self, item : SequenceItem = None) -> SequenceItem:
        """
        Get the next sequence item.

        The implementation ensures items are driven on the rising edge of pclk, when not in reset,
        while allowing for back-to-back requests if the sequencer provides them.

        :param item: The sequence item to retrieve, defaults to None
        :type item: SequenceItem, optional
        :return: The next sequence item
        :rtype: SequenceItem
        :raises NotImplementedError: If the method is not implemented in subclasses
        """

        while True:
            await RisingEdge(self.i_f.pclk)
            if self.i_f.get("presetn") == 0:
                continue

            if bool(self.i_f.get("psel") & (1 << self.idx)) and bool(self.i_f.get("pwakeup", 1)):

                next_item = SequenceItem(f"from_{self.name}", self)
                next_item.set("paddr", self.i_f.get("paddr"))
                next_item.set("psel", self.i_f.get("psel"))
                next_item.set("pwrite", self.i_f.get("pwrite"))
                next_item.set("pwdata", self.i_f.get("pwdata"))
                next_item.set("prdata", 0)
                next_item.set("pslverr", 0)
                next_item.set("pstrb", self.i_f.get("pstrb"))
                next_item.set("pprot", self.i_f.get("pprot"))
                next_item.set("pnse", self.i_f.get("pnse"))
                next_item.set("pauser", self.i_f.get("pauser"))
                next_item.set("pwuser", self.i_f.get("pwuser"))
                next_item.set("pruser", 0)
                next_item.set("pbuser", 0)

                return next_item


class CplRandomDriver(CplDriver):
    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the Random Driver for the APB agent.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

    async def get_next_item(self, item : SequenceItem = None) -> SequenceItem:
        """
        Get the next sequence item with randomization.

        In the random driver implementation, the item fields are randomized
        based on constraints from the sequence item.
        """
        item = await super().get_next_item(item)
        item.randomize_completion()

        return item

class CplMemoryDriver(CplDriver):
    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the Memory Driver for the APB agent.

        This driver uses a memory model to simulate the behavior of the APB interface.

        It allows for reading and writing to a memory model, which is useful for
        simulating memory-mapped devices.

        By default accesses which miss the allocated range will return slverr (if supported)
        and randomize the item.

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)
        self.memory = avl.Memory(width=self.i_f.DATA_WIDTH)
        self.memory.miss = lambda address : None

        # Add ranges to the memory if specified in the configuration
        self.ranges = avl.Factory.get_variable(f"{self.get_full_name()}.ranges", None)
        if self.ranges is not None:
            for r in self.ranges:
                self.memory.add_range(r[0], r[1])

    async def get_next_item(self, item : SequenceItem = None) -> SequenceItem:
        """
        Get the next sequence item as if targeting a memory.

        In the memory driver implementation, the item memory is updated and
        returns data from the memory if the address is valid.

        In the event the address is not valid, the item is randomized and a plslverr
        signal is set to indicate an error (if present)
        """
        item = await super().get_next_item(item)

        if self.memory._check_address_(item.get("paddr")):
            if item.get("pwrite", False):
                self.memory.write(item.get("paddr"), item.get("pwdata"), strobe=item.get("pstrb", None))
            else:
                item.set("prdata", self.memory.read(item.get("paddr")))
        else:
            if hasattr(item, "pslverr"):
                item.add_constraint("_c_pslverr_", lambda x: x == 1, item.pslverr)
            item.randomize_completion()
        return item

__all__ = ["CplDriver", "CplRandomDriver", "CplMemoryDriver"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Agent Configuration

import avl


class AgentCfg(avl.Object):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the avl-apb Agent Configuration

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        # Agent Attributes
        self.has_requester = avl.Factory.get_variable(f"{self.get_full_name()}.has_requester", False)
        """Has Requester Driver"""
        self.num_completer = avl.Factory.get_variable(f"{self.get_full_name()}.num_completer", 0)
        """Number Completer Driver"""
        self.has_monitor = avl.Factory.get_variable(f"{self.get_full_name()}.has_monitor", False)
        """Has Monitor Driver"""
        self.has_coverage = avl.Factory.get_variable(f"{self.get_full_name()}.has_coverage", False)
        """Has Functional Coverage"""
        self.has_bandwidth = avl.Factory.get_variable(f"{self.get_full_name()}.has_bandwidth", False)
        """Has Bandwidth Monitor"""
        self.has_trace = avl.Factory.get_variable(f"{self.get_full_name()}.has_trace", False)
        """Has Trace Generator"""

__all__ = ["AgentCfg"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Coverage

import avl

from ._item import SequenceItem


class Coverage(avl.Component):

    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize Coverage

        :param name: Name of the coverage class.
        :type name: str
        :param parent: Parent component.
        :type parent: Component
        """
        super().__init__(name, parent)

        self.item_port = avl.List()
        self.item = SequenceItem("for_coverage", self)

        # Define coverage
        self.cg = avl.Covergroup("apb", self)
        self.cg.set_comment("APB Coverage")

        # PSEL
        self.cp_psel = self.cg.add_coverpoint("psel", lambda: self.item.psel)
        self.cp_psel.set_comment("PSEL (1-hot select)")
        for i in range(self.item.psel.width):
            self.cp_psel.add_bin(f"{i}", lambda x, y=i : x & (1<<y))

        # PADDR
        self.cp_paddr_bits = self.cg.add_coverpoint("paddr_bits", lambda: self.item.paddr)
        self.cg.set_comment("PADDR (address bits)")
        for i in range(self.item.paddr.width):
            self.cp_paddr_bits.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
            self.cp_paddr_bits.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

        # PWRITE
        self.cp_pwrite = self.cg.add_coverpoint("pwrite", lambda: self.item.pwrite)
        self.cp_pwrite.set_comment("PWRITE (write enable)")
        for i in range(2):
            self.cp_pwrite.add_bin(f"{i}", i)

        # PWDATA
        self.cp_pwdata = self.cg.add_coverpoint("pwdata", lambda: self.item.pwdata if self.item.pwrite == 1 else 0)
        self.cp_pwdata.set_comment("PWDATA (write data)")
        for i in range(self.item.pwdata.width):
            self.cp_pwdata.add_bin(f"[{i}] == 0", lambda x ,y=i: 0 == (x & (1<<y)))
            self.cp_pwdata.add_bin(f"[{i}] == 1", lambda x ,y=i: 0 != (x & (1<<y)))

        # PRDATA
        self.cp_prdata = self.cg.add_coverpoint("prdata", lambda: self.item.prdata if self.item.pwrite == 0 else 0)
        self.cp_prdata.set_comment("PRDATA (read data)")
        for i in range(self.item.prdata.width):
            self.cp_prdata.add_bin(f"[{i}] == 0", lambda x ,y=i: 0 == (x & (1<<y)))
            self.cp_prdata.add_bin(f"[{i}] == 1", lambda x ,y=i: 0 != (x & (1<<y)))

        if hasattr(self.item, "wait_cycles"):
            # PREADY
            self.cp_wait_cycles = self.cg.add_coverpoint("wait_cycles", lambda: self.item.wait_cycles)
            self.cp_wait_cycles.set_comment("Wait cycles (number of cycles before PREADY after PENABLE)")
            for i in range(3):
                self.cp_wait_cycles.add_bin(f"{i}", i)
            self.cp_wait_cycles.add_bin("wait_cycles", range(0,1024), stats=True)

        if hasattr(self.item, "pslverr"):
            # PSLVERR
            self.cp_psvlverr = self.cg.add_coverpoint("pslverr", lambda: self.item.pslverr)
            for i in range(2):
                self.cp_psvlverr.add_bin(f"{i}", i)

            self.cc_pslverrXpwrite = self.cg.add_covercross("pslverrXpwrite", self.cp_psvlverr, self.cp_pwrite)
            self.cc_pslverrXpwrite.set_comment("Cross PSLVERR and PWRITE")

        if hasattr(self.item, "pstrb"):
            # PSTRB
            self.cp_pstrb = self.cg.add_coverpoint("pstrb", lambda: self.item.pstrb if self.item.pwrite == 1 else 0)
            self.cp_pstrb.set_comment("PSTRB (write strobe)")
            for i in range(self.item.pstrb.width):
                self.cp_pstrb.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
                self.cp_pstrb.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

            self.cp_pstrb_size = self.cg.add_coverpoint("pstrb_size", lambda: int(self.item.pstrb).bit_count() if self.item.pwrite == 1 else 0)
            self.cp_pstrb_size.set_comment("PSTRB (write strobe size)")
            for i in range(1, self.item.pstrb.width+1):
                self.cp_pstrb_size.add_bin(f"{i}", i)

        if hasattr(self.item, "pprot"):
            # PPROT
            self.cp_pprot = self.cg.add_coverpoint("pprot", lambda: self.item.pprot)
            self.cp_pprot.set_comment("PPROT (protection)")
            for i in range(8):
                self.cp_pprot.add_bin(f"{i}", i)

            self.cc_pprotXpwrite = self.cg.add_covercross("pprotXpwrite", self.cp_pprot, self.cp_pwrite)
            self.cc_pprotXpwrite.set_comment("Cross PPROT and PWRITE")

        if hasattr(self.item, "pnse"):
            # PNSE
            self.cp_pnse = self.cg.add_coverpoint("pnse", lambda: self.item.pnse)
            self.cp_pnse.set_comment("PNSE (non-secure enable)")
            for i in range(2):
                self.cp_pnse.add_bin(f"{i}", i)

            self.cc_pprotXpnse = self.cg.add_covercross("pprotXpnse", self.cp_pprot, self.cp_pnse)
            self.cc_pprotXpnse.set_comment("Cross PPROT and PNSE")

            self.cc_pnseXpwrite = self.cg.add_covercross("pnseXpwrite", self.cp_pnse, self.cp_pwrite)
            self.cc_pnseXpwrite.set_comment("Cross PNSE and PWRITE")

        if hasattr(self.item, "time_since_wakeup"):
            # PWAKEUP
            self.cp_pwakeup = self.cg.add_coverpoint("pwakeup", lambda: self.item.time_since_wakeup)
            self.cp_pwakeup.set_comment("PWAKEUP (wakeup indication - time in ns since wakeup was raised)")
            self.cp_pwakeup.add_bin("ns", range(0, 1000), stats=True)

        if hasattr(self.item, "pauser"):
            # PAUSER
            self.cp_pauser = self.cg.add_coverpoint("pauser", lambda: self.item.pauser)
            self.cp_pauser.set_comment("PAUSER (user request sideband)")
            for i in range(self.item.pauser.width):
                self.cp_pauser.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
                self.cp_pauser.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

        if hasattr(self.item, "pwuser"):
            # PWUSER
            self.cp_pwuser = self.cg.add_coverpoint("pwuser", lambda: self.item.pwuser)
            self.cp_pwuser.set_comment("PWUSER (user write sideband)")
            for i in range(self.item.pwuser.width):
                self.cp_pwuser.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
                self.cp_pwuser.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

        if hasattr(self.item, "pruser"):
            # PRUSER
            self.cp_pruser = self.cg.add_coverpoint("pruser", lambda: self.item.pruser)
            self.cp_pruser.set_comment("PRUSER (user read sideband)")
            for i in range(self.item.pruser.width):
                self.cp_pruser.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
                self.cp_pruser.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

        if hasattr(self.item, "pbuser"):
            # PBUSER
            self.cp_pbuser = self.cg.add_coverpoint("pbuser", lambda: self.item.pbuser)
            self.cp_pbuser.set_comment("PBUSER (user response sideband)")
            for i in range(self.item.pbuser.width):
                self.cp_pbuser.add_bin(f"[{i}] == 0", lambda x, y=i : 0 == (x & (1<<y)))
                self.cp_pbuser.add_bin(f"[{i}] == 1", lambda x, y=i : 0 != (x & (1<<y)))

    async def run_phase(self) -> None:
        """
        Run phase for the coverage component.

        """

        while True:
            # Wait for an item to be available
            self.item = await self.item_port.blocking_get()

            # Sample
            self.cg.sample()

__all__ = ["Coverage"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Agent

import avl
from cocotb.handle import HierarchyObject
from cocotb.triggers import NextTimeStep, RisingEdge

from ._agent_cfg import AgentCfg
from ._bandwidth import Bandwidth
from ._cdriver import CplDriver
from ._coverage import Coverage
from ._interface import Interface
from ._monitor import Monitor
from ._rdriver import ReqDriver
from ._rsequence import ReqSequence


class Agent(avl.Agent):
    def __init__(self, name: str, parent: avl.Component) -> None:
        """
        Initialize the avl-apb Agent

        :param name: Name of the agent instance
        :type name: str
        :param parent: Parent component
        :type parent: Component
        """
        super().__init__(name, parent)

        # Create configuration and export to children
        self.cfg = avl.Factory.get_variable(f"{self.get_full_name()}.cfg", AgentCfg("cfg", self))
        avl.Factory.set_variable(f"{self.get_full_name()}.*.cfg", self.cfg)

        # Bind HDL to establish parameters and configuration
        self._bind_(avl.Factory.get_variable(f"{self.get_full_name()}.hdl", None))

        # Create sequencer and driver if requester is enabled
        if self.cfg.has_requester:
            self.rsqr = avl.Sequencer("rsqr", self)
            self.rseq = ReqSequence("rseq",self.rsqr)
            self.rdrv = ReqDriver("rdrv", self)
            self.rsqr.seq_item_export.connect(self.rdrv.seq_item_port)

        if self.cfg.num_completer > 1:
            self.cdrv = []
            for i in range(self.cfg.num_completer):
                avl.Factory.set_variable(f"{self.get_full_name()}.cdrv_{i}.idx", i)
                self.cdrv.append(CplDriver(f"cdrv_{i}", self))
        elif self.cfg.num_completer == 1:
            self.cdrv = CplDriver("cdrv", self)

        # Create monitor if enabled
        if self.cfg.has_monitor:
            self.monitor = Monitor("monitor", self)

            if self.cfg.has_coverage:
                self.coverage = Coverage("coverage", self)
                self.monitor.item_export.connect(self.coverage.item_port)

            if self.cfg.has_bandwidth:
                self.bandwidth = Bandwidth("bandwidth", self)
                self.monitor.item_export.connect(self.bandwidth.item_port)

            if self.cfg.has_trace:
                self.trace = avl.Trace("trace", self)
                self.monitor.item_export.connect(self.trace.item_port)

    def _bind_(self, hdl) -> None:
        """
        Bind the agent to a hardware description language (HDL) interface.
        This method is used to associate the agent with a specific HDL interface,
        allowing it to interact with the hardware model.

        :param hdl: The HDL interface to bind to the agent
        :type hdl: HierarchyObject
        :raises TypeError: If `hdl` is not an instance of HierarchyObject
        """
        if not isinstance(hdl, HierarchyObject):
            raise TypeError(f"Expected HierarchyObject, got {type(hdl)}")

        # Assign Interface
        self.i_f = Interface(hdl)
        avl.Factory.set_variable(f"{self.get_full_name()}.*.i_f", self.i_f)

    async def run_phase(self) -> None:
        """
        Run the agent's phase. This method is called to start the agent's operation.
        It initializes the agent and starts the requester and completer if they are active.
        """

        self.raise_objection()
        await NextTimeStep()

        if self.cfg.has_requester:
            await self.rseq.start()

        # Run-off
        for _ in range(10):
            await RisingEdge(self.i_f.pclk)

        self.drop_objection()

__all__ = ["Agent"]



# Copyright 2025 Apheleia
#
# Description:
# Apheleia Verification Library Interface

from typing import Any

from cocotb.handle import HierarchyObject


class Interface:
    def __init__(self, hdl : HierarchyObject) -> None:
        """
        Create an interface
        Work around simulator specific issues with accessing signals inside generates.
        """
        # Parameters and Signals
        for child in list(hdl):
            # Parameters start with a capital letter
            if child._name[0].isupper():
                if isinstance(child.value, bytes):
                    setattr(self, child._name, str(child.value.decode("utf-8")))
                else:
                    setattr(self, child._name, int(child.value))
            else:
                setattr(self, child._name, child)

        if self.CLASSIFICATION != "APB":
            raise TypeError(f"Expected APB classification, got {self.CLASSIFICATION}")

        if self.VERSION not in [2, 3, 4, 5]:
            raise ValueError(f"Unsupported APB version: {self.VERSION}")

        if self.PSEL_WIDTH < 1:
            raise ValueError(f"Invalid PSEL_WIDTH: {self.PSEL_WIDTH}")

        # Remove un-configured signals
        if self.VERSION < 3:
            delattr(self, "pready")
            delattr(self, "pslverr")

        if self.VERSION < 4 or self.Protection_Support == 0:
            delattr(self, "pprot")

        if self.VERSION < 4 or self.Pstrb_Support == 0:
            delattr(self, "pstrb")

        if self.VERSION < 5 or self.RME_Support == 0:
            delattr(self, "pnse")

        if self.VERSION < 5 or self.Wakeup_Signal == 0:
            delattr(self, "pwakeup")

        if self.VERSION < 5 or self.USER_REQ_WIDTH == 0:
            delattr(self, "pauser")

        if self.VERSION < 5 or self.USER_DATA_WIDTH == 0:
            delattr(self, "pwuser")
            delattr(self, "pruser")

        if self.VERSION < 5 or self.USER_RESP_WIDTH == 0:
            delattr(self, "pbuser")

    def set(self, name : str, value : int) -> None:
        """
        Set the value of a signal (if signal exists)

        :param name: The name of the signal
        :type name: str
        :param value: The value to set
        :type value: int
        :return: None
        """
        signal = getattr(self, name, None)
        if signal is not None:
            signal.value = value

    def get(self, name : str, default : Any = None) -> int:
        """
        Get the value of a signal (if signal exists)

        :param name: The name of the signal
        :type name: str
        :param default: The default value to return if signal does not exist
        :type default: Any
        :return: The value of the signal or the default value
        :rtype: int
        """
        signal = getattr(self, name, None)
        if signal is not None:
            return int(signal.value)
        return default

__all__ = ["Interface"]
